/**
 * Copyright (c) 2007 University of Patras Computer Engineering and Informatics Department Greece
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
  */

#include <stdlib.h>
#include "packet.h"
#include "ip.h"
#include "rtp.h"
#include <math.h>
#include "simstream.h"

// Default values.. you can change them for your experiments
#define CONGESTED 0.2
#define UNLOADED  0.9

#define NOW Scheduler::instance().clock()
#define TIMER_TIMEOUT 10
#define INITIAL_TIMEOUT 10



/** creates an instance of the RTPSource class
   returns the newly created object
*/
static class RTPSourceClass : public TclClass
{
	public:
		RTPSourceClass() : TclClass ( "RTPSource" ) {}
		TclObject* create ( int argc, const char*const* argv )
		{
			if ( argc >= 5 )
				return ( new RTPSource ( atoi ( argv[4] ) ) );

			return 0;
		}
} class_rtp_source;

/**
 *
 * creates a SessionRTP instance
 *
*/
static class RTPSessionClass : public TclClass
{
	public:
		RTPSessionClass() : TclClass ( "Session/RTP" ) {}
		TclObject* create ( int, const char*const* )
		{
			return ( new RTPSession() );
		}
} class_rtp_session;

/**
* zero argument constructor
*/
RTPSession::RTPSession()
		: allsrcs_ ( 0 ), localsrc_ ( 0 ), receivers_ ( 0 ),stream_ ( 0 ),last_np_ ( 0 ),sender_timer_ ( this ), T_one_way_ ( 0 ),rh_ ( 0 ),we_sent ( 0 ),last_pkts_lost_ ( 0 ),last_ehsr_ ( -1 ),tx_rate_ ( 0 ), time_elapsed_ ( 0 ),running_ ( 0 )
{
	bind ( "enableFlowControl_", &enableFlowControl_ );
//	bind ( "rx_recv_",&rx_recv_ );
	bind ( "RTT_",&RTT_ );
	bind ( "out_jitter_",&out_jitter_ );
	bind ( "smooth_loss_",&smooth_loss_ );
	bind ( "level_",&level_ );
	bind ( "cjitter_", &cjitter_ );
	ts_p= ts_n= tr_p= tr_n=0;
	jitter_= 0;
	nlong_samples =  nshort_samples = 0;
	long_avg_cjitter =  short_avg_cjitter = 0;
	long_sum =  short_sum = 0;
	smooth_rate_ = 0;
	alpha = 1;
	smooth_factor = 0.1;
	last_time_report_ = NOW;
	low_ = 0;
	high_= 0;
	avg_tx_ = 0;
	

	weight[0] = 1;
	weight[1] = 1;
	weight[2] = 1;
	weight[3] = 1;
	weight[4] = 0.8;
	weight[5] = 0.6;
	weight[6] = 0.4;
	weight[7] = 0.2;
}

/**
 * destructor
*/
RTPSession::~RTPSession()
{
	while ( allsrcs_ != 0 )
	{
		RTPSource* p = allsrcs_;
		allsrcs_ = allsrcs_->next;
		delete p;
	}
	delete localsrc_;
	while ( receivers_ != 0 )
	{
		RTPReceiver* p = receivers_;
		receivers_ = receivers_->next;
		delete p;
	}
	lst.~list<double>();
	tx_lst.~list<double>();

}

/**
 * it is called when the RTPAgent sends a RTP packet
   and updates the number of packets (np) field of the
   originator of the RTP packet
 * @param int the packet to be added
 */
void RTPSession::localsrc_update ( int )
{
	localsrc_->np ( 1 );
	localsrc_->is_sender ( true );
}

/**
 *  it is called when the RTPAgent sends a RTP packet
 *  and updates the number of number of bytes (nbytes) field of the
 *  originator of the RTP packet
 *  @param n the bytes to be added
 */
void RTPSession::localsrc_update_nbytes ( int n )
{
	localsrc_->nbytes ( n );

}

#define RTCP_HDRSIZE 8
#define RTCP_SR_SIZE 20
#define RTCP_RR_SIZE 48

/**
 *  the method for creating the SR and RR reports
 *  the "heart" of the RTCP protocol
 * @param bye if there is a bye event
 * @return the length of this report in bytes
 */
int RTPSession::build_report ( int bye )
{
//printf("INSIDE BUILD REPORT TIME: %f Local %d  \n", NOW,localsrc_->srcid());

	rh_ = new hdr_rtp;
	rh_->rr_ = 0;
	rh_->sr_ = 0;

	int nsrc = 0;
	int nrr = 0;
	int len = RTCP_HDRSIZE;
	we_sent = 0;
	double fraction = 0.0;
	time_elapsed_ = NOW - last_time_report_;
	last_time_report_ = NOW;
	if ( localsrc_->np() != last_np_ )
	{
		last_np_ = localsrc_->np();
		we_sent = 1;
		len += RTCP_SR_SIZE;
		//add sender report
		sender_report* sr;
		//fill the report
		sr = new sender_report;
		sr->s_ext_ = 0;
		sr->sender_srcid() = localsrc_->srcid();
		sr->pkts_sent() = localsrc_->np();
		sr->octets_sent() = localsrc_->nbytes();
		sr->low_tx() =stream_->getLow_tx();
		sr->medium_tx() =stream_->getMedium_tx();
		sr->high_tx() =stream_->getHigh_tx();
		RTPReceiver *p;
		for ( p = receivers_; p != 0; p = p->next )
		{
			sr_extensions* se = new sr_extensions;
			se->srcid() =p->srcid();
			se->eff_rtt() =p->eff_rtt();
			se->next = sr->s_ext_;
			sr->s_ext_ = se;
		}
		//store the report
		rh_->sr_ = sr;
	}
	for ( RTPSource* sp = allsrcs_; sp != 0; sp = sp->next )
	{
		++nsrc;
		int received = sp->np() - sp->snp();
		stream_->set_total_pkts_recv(received);

		if ( received == 0 )
		{

			continue;
		}
		if ( localsrc_->srcid() != sp->srcid() )
		{

//			int bytes = received * ( sp->ps() );
//			rx_recv_ = ( 8 * ( double ) bytes ) /time_elapsed_;
//			stream_->setRcv_rate ( rx_recv_ );
			sp->snp ( sp->np() );
			len += RTCP_RR_SIZE;
			// calculate loss fraction since previous report
			int expected_interval  = sp->ehsr() - last_ehsr_;
			last_ehsr_ = sp->ehsr();
			int lost_interval = expected_interval -  received;
			stream_->set_total_pkts_lost(lost_interval);

			if ( lost_interval <= 0 || expected_interval == 0 )
			{
				fraction = 0;

			}
			else fraction = ( ( double ) lost_interval  / ( double ) expected_interval );
			calculate_RTT();
			// Update the R_tcp according to fraction value
			if ( sp->np() >= 1 )  // I have already received RTP packets from the source
			{
				if ( fraction == 0 )
				{
					increase_rate ( sp->ps() );

				}
				else /*if(fraction != 0)*/
				{
					measure_smooth_loss ( fraction );
					calculateR_tcp ( sp->ps() );
				}
			}
			//add receiver report
			receiver_report* rr;
			rr = new receiver_report;
			//fill the report
			rr->cum_pkts_lost() = sp->cum_pkts_lost();
			rr->LSR() = sp->LSR();
			rr->DLSR() = Scheduler::instance().clock() - sp->SRT();
			rr->R_tcp() = smooth_rate_;
			rr->jitter() = sp->jitter();
			rr->rx() =stream_->getAvg_rx();
			avg_cjitter();
			if ( bye )
			{
				rr->bye() = 1;
				remove_sender ( sp );
			}
			else rr->bye() = 0;

			rh_->rr_ = rr;

			if ( ++nrr >= 31 )
				break;
		} // end if

	}// end for loop


	if ( bye )
	{
		len += build_bye();
	}
	else
		len += build_sdes();

	Tcl::instance().evalf ( "%s adapt-timer %d %d %d", name(),
	                        nsrc, nrr, we_sent );
	Tcl::instance().evalf ( "%s sample-size %d", name(), len );

	return ( len );
}

/** it is called upon a bye event
   returns 8 bytes length
*/
int RTPSession::build_bye()
{

	return ( 8 );
}
/** not implemented yet. It only returns
   20 bytes that are to be added to the RTCP report
*/
int RTPSession::build_sdes()
{

	/* XXX We'll get to this later... */
	return ( 20 );
}

/**
 *  this method accesses the RTP packets
 *  it adds a newly heard source, which is the originator of the RTP packet,
 *  calculates the one way trip time and calls the calculate_RTT(T_one_way_ , eff_rtt)
 *  method for the RTT calculation.
 *  It measures also the delay jitter time of the RTP packets.
 *  Use this method to update any wished field of the Source class instances
 * @param p  the RTP packet that just received by the RTP Agent
 * @param
 */
void RTPSession::recv ( Packet* p, Handler* )
{

	hdr_cmn* mh = hdr_cmn::access ( p );
	hdr_rtp* rh = hdr_rtp::access ( p );
	RTPSource* s = lookup ( rh->srcid() );

	if ( s == 0 )
	{
		if ( rh->srcid() !=localsrc_->srcid() )
		{
			Tcl& tcl = Tcl::instance();
			tcl.evalf ( "%s new-source %d", name(),rh->srcid() );
			s = ( RTPSource* ) TclObject::lookup ( tcl.result() );
			last_ehsr_ = rh->seqno();
		}
	}
	if ( rh->srcid() !=localsrc_->srcid() )
	{
		// calculate one way trip time
		T_one_way_ = NOW - rh->timestamp();
		// count packet losses
		int pkts_lost = 0;
		int difference =  rh->seqno() - s->ehsr();
		if ( difference <= 1 ) // to deal with the first RTP packet in which s->ehrs() = -1
		{
			pkts_lost = 0;

		}

		if ( difference > 1 )
		{
			pkts_lost = difference -1;

		}
		// calculate delay jitter, see RFC 3550 Appendix A8
		double transit = NOW - rh->timestamp();
		//printf("transit %f \n", transit);
		// end-to-ende delay
		stream_->set_e2e_delay(transit);
		double d = transit - s->transit();
		s->transit ( transit );
		if ( d < 0 ) d = -d;
		s->jitter ( s->jitter() + ( 1./16. ) * ( d - s->jitter() ) );
		stream_->set_jitter(s->jitter());
		
		// calculate interframe jitter
		ts_p = ts_n;
		ts_n = rh->timestamp();
		tr_p = tr_n;
		tr_n = NOW;
		out_jitter_ = ( tr_n-tr_p )- ( ts_n-ts_p );
		// this is the cumulative jitter
		cjitter_+= out_jitter_;
		// we use it for the averaging functions
		sum_cjitter ( cjitter_ );
//              printf("jitter %f \n",s->jitter());
		s->np ( 1 ); // counts the packets that I have received from sender s
		s->cum_pkts_lost ( pkts_lost );// counts the packets that I have lost from sender s
		s->ehsr ( rh->seqno() );// the first ehsr = -1
		s->nbytes ( mh->size() ); // counts the number of bytes that I have received from sender s
		s->ps ( mh->size() );// the packet size in bytes
		stream_->setRcv_bytes(mh->size());
		
		
		if ( rh->T_epoch() == 1 )
		{
			stream_->compare();
		}
	}
	Packet::free ( p );
}

/**
 * We sum up the measured cumulative jitter and keep a moving long and short
 * average.
 * @param value  the new value of cumulative jitter
 */
void RTPSession::sum_cjitter ( double value )
{
	nlong_samples++;
	nshort_samples++;
	long_avg_cjitter = long_avg_cjitter + ( ( value - long_avg_cjitter ) /nlong_samples );
	short_sum += value;
//	printf("Time %f value %f samples %d AVG-CJITTER %f \n", NOW,value,nlong_samples, long_avg_cjitter );

}


/**
 * We compare the long and short averages. When the short average
 * is bigger than the long average it is an indication of upcoming
 * congestion. The CONGESTED factor takes a low value to smooth
 * the transmission rate of the sender.When the network is not congested we
 * ask from the sender to increase the transmission rate.
 */
void RTPSession::avg_cjitter()
{
	if ( nshort_samples >0 )
	{
		short_avg_cjitter = short_sum/ ( double ) nshort_samples;
	}
	if ( short_avg_cjitter >= long_avg_cjitter )
	{
		smooth_factor = CONGESTED;
//	printf("TIME-CONGESTED %f localsrc_  %d  short_avg_cjitter %f  long_avg_cjitter %f \n", NOW, localsrc_->srcid(),short_avg_cjitter, long_avg_cjitter  );
	}
	else
	{
		smooth_factor = UNLOADED;
	}

	nshort_samples = 0;
	short_sum = 0;


}

/**
 *  the function that calculate the A value
 * @param value the effective RTT that is measured by the RTP sender
 */
void RTPSession::calculate_alpha ( double value )
{

	if ( value == 0 )
	{
		alpha = 1;
	}
	else
		alpha = ( value/T_one_way_ ) -1;
}


/**
 *  calculates the RTT time
 *  one_way is the one way trip time
 *  eff is the effective RTT time measured and reported by the sender of the
 *  RTP packets
 * @param eff, the effective RTT time measured by the sender
 */
void RTPSession::calculate_RTT()
{
	double temp = ( 1+alpha ) * T_one_way_;
//	RTT_ = temp * 0.5 + 0.5 * RTT_;
	RTT_ = temp * 0.1 + 0.9 * RTT_;   //taken from RFC 3448
}

/**
 * the estimation made by the receiver of the transmission rate when
 *  fraction loss is zero. The increase is not higher than 1 RTP packet/current rate
 *  in bytes/sec
 *  sample_tx is the current estimation of the RTP receiver. We pass this value from
 *  the filter
 * @param ps, the packet size of the RTP packet
 */
void RTPSession::increase_rate ( int ps )
{
	double sample_tx = smooth_rate_ + 3*( double ) ps/RTT_;
	// pass sample_tx from the filter
	smooth_rate ( sample_tx );
}

/**
 * measure_smooth_loss is called when the last measured by the receiver
 * loss fraction is not zero.
 * @param fraction  the fraction of lost RTP packets/expected RTP packets
 * since last RR
 */
void RTPSession::measure_smooth_loss ( double fraction )
{

	double I_tot0 = 0;
	double I_tot1 = 0;
	double W_tot = 0;
	for ( int i=0; i<7; i++ )
	{
		pkt_loss_history[i+1]= pkt_loss_history[i];
	}
	pkt_loss_history[0] = fraction;

	for ( int i=0; i<7; i++ )
	{
		I_tot0 = I_tot0+ ( pkt_loss_history[i] * weight[i] );
		W_tot = W_tot +  weight[i];
	}

	for ( int i=1; i<8; i++ )
	{
		I_tot1 = I_tot1 + ( pkt_loss_history[i] * weight[i-1] );
	}
	double I_tot =0;
	if ( I_tot0 > I_tot1 )
	{
		I_tot = I_tot0;
	}
	else
	{
		I_tot = I_tot1;
	}
	double  I_mean = I_tot/W_tot;

	smooth_loss_= I_mean;
}

/**
 *  calculates the TCP friendly bandwidth share
 * @param ps is the size of the RTP packet
 */
void RTPSession::calculateR_tcp ( int ps )
{

	double min =0;
	double fraction = 3 * sqrt ( ( 3 * smooth_loss_ ) / 8 );
	if ( fraction <  1 )
	{
		min = fraction;
	}
	else min =1;
	// TCP friendly BW share equation
	double sample_tx = ( double ) ps / ( RTT_ * ( sqrt ( 2 * smooth_loss_ /3 ) ) + 4 * ( RTT_ * min * smooth_loss_ ) * ( 1 + ( 32 * pow ( smooth_loss_,2 ) ) ) );
	// pass sample_tx from the filter
	smooth_rate ( sample_tx );
}



/**
 * this is the filter function for the smooth transmission rate measurement
 * @param inst_tx  the current transmission rate estimation by the RTP receiver
 */
void RTPSession::smooth_rate ( double inst_tx )
{
	smooth_rate_ = smooth_factor * inst_tx + smooth_rate_ * ( 1 - smooth_factor );
	stream_->setInst_rx_rate ( smooth_rate_ );
}



/**
 * the function that accesses and exploits the SR and RR
 *  here is where the sender calculates the effective RTT time based on
 *  the receiver's RR. The sender also updates its transmission rate based
 *  on the receivers' RR.
 *  We don't exploit the SR in this implementation but a third party eg
 *  a network controller can exploit both SR and RR, witout in fact joining the session
 * @param p the RTCP packet
 */
void RTPSession::recv_ctrl ( Packet* p )
{
	hdr_cmn* mh = hdr_cmn::access ( p );
	hdr_rtp* rh = hdr_rtp::access ( p );
	u_int32_t local_src = localsrc_->srcid();

	if ( rh->srcid() != local_src )
	{

		if ( rh->sr_ != 0 ) // get the Sender report
		{
			RTPSource* source = lookup ( rh->sr_->sender_srcid() );
			if ( source != 0 && source->srcid() != local_src )
			{
				source->LSR ( rh->timestamp() );
				source->SRT ( NOW );
				stream_->setLow_tx ( rh->sr_->low_tx() );
				stream_->setMedium_tx ( rh->sr_->medium_tx() );
				stream_->setHigh_tx ( rh->sr_->high_tx() );
				if ( rh->sr_->s_ext_ != 0 )
				{
					sr_extensions* se = new sr_extensions;
					for ( se = rh->sr_->s_ext_; se != 0; se = se->next )
					{

						if ( se->srcid() ==localsrc_->srcid() )
						{

							if ( se->eff_rtt() == 0 )
							{
								calculate_alpha ( 0 );
							}
							else calculate_alpha ( se->eff_rtt() );
						}

					}// end for
				}

			}
		}

		if ( localsrc_->is_sender() )   // get the Receiver report
		{

			if ( rh->rr_ != 0 )
			{
				RTPReceiver* s = lookup_rcv ( rh->srcid() );

				if ( s != 0 )
				{
					for ( s = receivers_; s!= 0; s = s->next )
					{

						if ( rh->srcid() == s->srcid() )
						{

							double eff_rtt = 0.0;

							if ( rh->rr_->LSR() !=0 )
							{
								eff_rtt = NOW - rh->rr_->LSR() - rh->rr_->DLSR();
							}
							s->eff_rtt ( eff_rtt );
							s->rate ( rh->rr_->R_tcp() );
							if ( rh->rr_->bye() == 1 )
							{
								remove_receiver ( s );
							}

						}

					} // end for loop
					if ( enableFlowControl_ == 1 && receivers_ )
					{
						update_rate();
					}
					else
					{
						printf ( "TIME: %f Sender %d NOT HAVING RECEIVERS \n", NOW,localsrc_->srcid() );
						slow_start();

					}
				} //end if
				else
				{
					s = new RTPReceiver ( rh->srcid() );
					enter_rcv ( s );
				}

			}
		}
	}

	Tcl::instance().evalf ( "%s sample-size %d", name(), mh->size() );
	Packet::free ( p );
}

/** this method compares the reported TCP friendly bandwidth
   by the receivers and calls the session_bw and transmit methods
   in session-rtp.tcl class to adjust the sender's transmission rate
   The adjustment in the transmission rate occures after the sender
   receives reports from all the receivers.
   r_tcp, is the receiver estimation of the TCP friendly bandwidth
*/

void RTPSession::update_rate()
{
	RTPReceiver* s;
	for ( s = receivers_; s!= 0; s = s->next )
	{
		if ( s->rate() > 0 )
		{
			lst.push_front ( s->rate() );
		}
	}
	lst.sort();
	double inst_tx =lst.front();
	lst.clear();
	if ( inst_tx != tx_rate_ )
	{
 		tx_rate_ = inst_tx;
// 		if ( tx_rate_ <=  low_ )
// 		{
// 			tx_rate_ = low_;
// 		}
// 		if ( tx_rate_ > high_ )
// 		{
// 			tx_rate_ =  high_;
// 		}
		
		Tcl::instance().evalf ( "%s session_bw %fb/s",name(),8*tx_rate_ );
		Tcl::instance().evalf ( "%s transmit %fb/s",name(),8*tx_rate_ );
	}
}


/**
 *  we start a slow start when all receivers have left the session
 */
void RTPSession::slow_start()
{
	tx_rate_ = tx_rate_/2;
	
	printf ( "Starting Slow start at ...%f\n", tx_rate_ * 8 );
	Tcl::instance().evalf ( "%s session_bw %fb/s",name(),8*tx_rate_ );
	Tcl::instance().evalf ( "%s transmit %fb/s",name(),8*tx_rate_ );
}


/**
 * accesses the RTP packet header
 * @return , the
 */
hdr_rtp* RTPSession::access_hdr_rtp()
{
	return rh_;
}

/**
 * it checks if the source has heard
 * @param srcid the source id of this source
 * @return the source of the RTP packet if it exists,
 * zero otherwise
 */
RTPSource* RTPSession::lookup ( u_int32_t srcid )
{
	RTPSource *p;
	for ( p = allsrcs_; p != 0; p = p->next )
	{
		if ( p->srcid() == srcid )
			return ( p );
	}
	return ( 0 );
}

/**
 * Check if the receiver exists in the sender's list
 * @param srcid  the source id of this receiver
 * @return  the receiver if it exists,
 *  zero otherwise
 */
RTPReceiver* RTPSession::lookup_rcv ( u_int32_t srcid )
{

	RTPReceiver *p;
	for ( p = receivers_; p != 0; p = p->next )
	{
		if ( p->srcid() == srcid )
			return ( p );
	}
	return ( 0 );
}

/** for debugging, prints all the receivers
    that a source serves
*/
void RTPSession::print_rcvs( )
{

	RTPReceiver *p;
	for ( p = receivers_; p != 0; p = p->next )
	{
		printf ( "Receiver:%d\n", p->srcid() );
	}
}

/** for debugging, prints all the sources
    that a receiver listens
*/
void RTPSession::print_srcs( )
{

	RTPSource* s;
	for ( s = allsrcs_; s != 0; s = s->next )
	{
		printf ( "Source:%d\n", s->srcid() );
	}
}


/**
 * enter a new source
 * @param s the new RTPSource
 */
void RTPSession::enter ( RTPSource* s )
{

	s->next = allsrcs_;
	allsrcs_ = s;
}

/**
 *   enter a new receiver
 * @param s the new RTPReceiver
 */
void RTPSession::enter_rcv ( RTPReceiver* s )
{
	printf ( "Time:%f Sender:%d ...I add receiver %d \n", NOW,localsrc_->srcid(), s->srcid() );
	s->next = receivers_;
	receivers_ = s;

}

/**
 * remove a receiver from a sender's list
 * @param p the RTPReceiver to be removed
 */
void RTPSession::remove_receiver ( RTPReceiver* p )
{

	printf ( "Time:%f Sender:%d ...I remove receiver %d \n", NOW,localsrc_->srcid(),p->srcid() );

	RTPReceiver* s;
	RTPReceiver* q;
	if ( receivers_== p )
	{
		q = receivers_;
		receivers_= receivers_->next;
		delete q;
	}
	else
	{
		for ( s = receivers_; s!=0; s=s->next )
		{
			if ( s->next==p )
			{
				q = s->next;
				s->next=q->next;
				delete q;
			}
		}
	}

}

/**
 *  we remove this sender from the list upon a bye event
 * @param p  the sender that has to be removed from our sources' list
 */
void RTPSession::remove_sender ( RTPSource* p )
{

	if ( p->srcid() != localsrc_->srcid() )
	{
		RTPSource* s;
		RTPSource* q;
		if ( allsrcs_==p )
		{
			q = allsrcs_;
			allsrcs_= allsrcs_->next;
			delete q;
		}
		else
		{
			for ( s = allsrcs_; s!=0; s=s->next )
			{
				if ( s->next==p )
				{
					q = s->next;
					s->next=q->next;
					delete q;
				}
			}
		}

	}
}

/**
 * it's good for the receiver to have
   a rough knowledge of the session rate
 * @param a  the initial session rate
 */
void RTPSession::initial_rate ( double a )
{
	tx_rate_ = ( double ) a/8;
	smooth_rate_ = tx_rate_;


}

/**
  * the command function that handles TCL commands
 *
 */
int RTPSession::command ( int argc, const char*const* argv )
{
	if ( argc == 3 )
	{
		if ( strcmp ( argv[1], "enter" ) == 0 )
		{
			RTPSource* s = ( RTPSource* ) TclObject::lookup ( argv[2] );
			enter ( s );
			return ( TCL_OK );
		}
		if ( strcmp ( argv[1], "localsrc" ) == 0 )
		{
			localsrc_ = ( RTPSource* ) TclObject::lookup ( argv[2] );
			enter ( localsrc_ );
			return ( TCL_OK );
		}
		if ( strcmp ( argv[1], "rate" ) == 0 )
		{
			initial_rate ( atoi ( argv[2] ) );
			return ( TCL_OK );
		}
		if ( strcmp ( argv[1], "stream" ) == 0 )
		{
			stream_ = ( SimStream* ) TclObject::lookup ( argv[2] );
			return ( TCL_OK );
		}
	}
	if ( argc == 4 )
	{
		if ( strcmp ( argv[1], "bw-limits" ) == 0 )
		{
			low_ = atoi ( argv[2] ) /8;
			high_ = atoi ( argv[3] ) /8;
			return ( TCL_OK );
		}
	}

	return ( TclObject::command ( argc, argv ) );
}

/**
 * Constructor
 * @param srcid the source id of the newly created RTPReceiver
 */
RTPReceiver::RTPReceiver ( u_int32_t srcid )
		: next ( 0 ),cum_pkts_lost_ ( 0 ),eff_rtt_ ( 0 ),rate_ ( 0 ),sent_report_ ( false )
{
	srcid_ = srcid;
}


/**
 * Constructor of the RTPSource class
 * @param srcid the source id of the newly created RTPSource
 */
RTPSource::RTPSource ( u_int32_t srcid )
		: next ( 0 ),np_ ( 0 ), snp_ ( 0 ), ehsr_ ( -1 ), nbytes_ ( 0 ),cum_pkts_lost_ ( 0 ),LSR_ ( 0 ),SRT_ ( 0 ),is_sender_ ( false ),ps_ ( 0 ),rate_ ( 0 ),transit_ ( 0 ),jitter_ ( 0 )
{
	bind ( "srcid_", ( int* ) &srcid_ );
	srcid_ = srcid;
}


// Simulcast functions

void SenderTimer::expire ( Event* /*e*/ )
{
	a_->timeout ( 0 );
}

void RTPSession::start_timer()
{
	running_ = 1;
	sender_timer_.resched ( INITIAL_TIMEOUT );

}

void RTPSession::stop_timer()
{
	running_ = 0;
	sender_timer_.force_cancel();

}



void RTPSession::timeout ( int )
{
	if ( running_ )
	{
		average_tx();
		stream_->setAvg_tx ( avg_tx_,level_ );
		sender_timer_.resched ( TIMER_TIMEOUT );
	}
}

/**
 *  add the current transmission rate in the list
 * @param newValue  the new instantaneous transmission rate
 */
void RTPSession::setInst_tx ( int newValue )
{
	tx_lst.push_front ( newValue );
}



/**
 *  it averages the transmission rate over a period Δt (TIMER_TIMEOUT)
 */
void RTPSession::average_tx()
{
	
	avg_tx_ = (double)localsrc_->nbytes()/TIMER_TIMEOUT;
	localsrc_->reset_nbytes(0);
 	printf("#####TIME %f source %d avg_tx_ %f \n", NOW, localsrc_->srcid(),avg_tx_); 
	
}




